/******************************************************************************
*                                                                             *
*       Copyright (C) 1993,1994 Tony Robinson				      *
*                                                                             *
*       See the file SLICENSE for conditions on distribution and usage	      *
*                                                                             *
******************************************************************************/

# include <sys/types.h>
# include <stdio.h>
# include <math.h>
# include "Slib.h"

# define DEFAULT_HEADERSIZE	1024
# define DEFAULT_FRAMESTEP	256
# define DEFAULT_WINDOWSIZE	(2 * DEFAULT_FRAMESTEP)
# define DEFAULT_MINSILENCE	0.05
# define DEFAULT_NCEPSTRA	12
# define DEFAULT_MAXDEV		3.0

static int floatcompare(float *a, float *b) {
  return(*a > *b);
}

int main(argc, argv) int argc; char **argv; {
  int swapbytes;
  int headersize   = DEFAULT_HEADERSIZE;
  int framestep    = DEFAULT_FRAMESTEP;
  int windowsize   = DEFAULT_WINDOWSIZE;
  float minsilence = DEFAULT_MINSILENCE;
  int ncepstra     = DEFAULT_NCEPSTRA;
  float maxdev	   = DEFAULT_MAXDEV;
  FILE *filei = NULL;
  short *wave;
  float *window, *energy, *sortenergy;
  float elimit, **cepCache, *acMean, *acVar;
  int   nsample, frame, nframe, *cepCacheValid;
  int	i, nsum, nacoustic, initframe, quitframe;

  Sgetargs(argc, argv, "-h", "-help", NULL,
	"-swapbytes",	NULL, &swapbytes, "swap the byte order of the signals",
	"-headersize",	"%d", &headersize, "size of header in bytes",
	"-framestep",	"%d", &framestep, "analysis frame step",
	"-windowsize",	"%d", &windowsize, "analysis window size",
	"-minsilence","%5.3f",&minsilence,"minimum fraction of silence frames",
	"-ncepstra",  "%d",   &ncepstra, "number of cepstral coefficients",
	"-maxdev",   "%5.3f",&maxdev,"maximum deviation of speech from silece",
	NULL,
	"speech file", "r", &filei, NULL, "16 bit speech waveform file",
	NULL);

  /* read in the waveform */
  wave = Sfreadwave(filei, NULL, headersize, windowsize - framestep, &nsample);
  if(swapbytes)
    Sswab(wave, wave, 2 * (nsample + windowsize - framestep));

  nframe = nsample / framestep;
  window = Svector(windowsize, sizeof(*window));
  energy = Svector(nframe, sizeof(*energy));
  sortenergy = Svector(nframe, sizeof(*energy));

  cepCache = (float**) Smatrix(nframe, ncepstra, sizeof(**cepCache));
  cepCacheValid = Svector(nframe, sizeof(*cepCacheValid));
  for(frame = 0; frame < nframe; frame++)
    cepCacheValid[frame] = 0;

  for(frame = 0; frame < nframe; frame++) {
    SzeroMeanHammWindow(wave + frame * framestep, windowsize, window);
    Swav2acf(window, windowsize, energy + frame, 1);
  }

  for(frame = 0; frame < nframe; frame++)
    sortenergy[frame] = energy[frame];
  qsort(sortenergy, nframe, sizeof(*sortenergy), floatcompare);

  elimit = sortenergy[(int) (minsilence * nframe)];

  /* go through all frames computing cepstra for quiet windows */
  for(frame = 0; frame < nframe; frame++)
    if(energy[frame] < elimit) {
      int offset;

      for(offset = -1; offset <= 1; offset++) {
	int rframe = abs(frame + offset);
	if(rframe >= nframe) rframe = 2 * nframe - rframe - 2;

	if(!cepCacheValid[rframe]) {
	  SzeroMeanHammWindow(wave + rframe * framestep, windowsize, window);
	  Swav2cep(window, windowsize, cepCache[rframe], ncepstra);
	  cepCacheValid[rframe] = 1;
	}       
      }
    }

  /* grab space for acoustic means and variances */
  nacoustic = 2 * ncepstra;
  acMean = Svector(nacoustic, sizeof(*acMean));
  acVar  = Svector(nacoustic, sizeof(*acVar));

  for(i = 0; i < nacoustic; i++) {
    acMean[i] = 0.0;
    acVar[i]  = 0.0;
  }

  /* make a second pass computing cepstra and delta cepstra sums */
  /* assume a diagional covaariance matrix */
  /* ( forget the first and last frame ) */
  nsum = 0;
  for(frame = 1; frame < nframe - 1; frame++)
    if(energy[frame] < elimit) {
      for(i = 0; i < ncepstra; i++) {
	float val = cepCache[frame][i];
	acMean[i] += val;
	acVar[i]  += val * val;
      }
      for(i = 0; i < ncepstra; i++) {
	float val = cepCache[frame - 1][i] - cepCache[frame + 1][i];
	acMean[i + ncepstra] += val;
	acVar[i + ncepstra]  += val * val;
      }
      nsum++;
    }

  /* compute means and variances */
  for(i = 0; i < nacoustic; i++) {
    acMean[i] = acMean[i] / nsum;
    acVar[i]  = acVar[i] / nsum - acMean[i] * acMean[i];
  }

  {
    int done = 0;

    for(frame = 1; !done && frame < nframe - 1; frame++) {
      int offset;
      float sumvar = 0.0;

      /* make sure we have all the cepstral values */
      for(offset = -1; offset <= 1; offset++) {
	int rframe = abs(frame + offset);
	if(rframe >= nframe) rframe = 2 * nframe - rframe - 2;

	if(!cepCacheValid[rframe]) {
	  SzeroMeanHammWindow(wave + rframe * framestep, windowsize, window);
	  Swav2cep(window, windowsize, cepCache[rframe], ncepstra);
	  cepCacheValid[rframe] = 1;
	}       
      }
      
      for(i = 0; i < ncepstra; i++) {
	float val = cepCache[frame][i] - acMean[i] ;
	sumvar += val * val / acVar[i];
      }
      for(i = 0; i < ncepstra; i++) {
	float val = cepCache[frame - 1][i] - cepCache[frame + 1][i] -
	  acMean[i + ncepstra];
	sumvar += val * val / acVar[i + ncepstra];
      }
/*
      printf("%d\t%f\t%f\n", frame * framestep, energy[frame], sumvar);
*/
      done = (sumvar / nacoustic > maxdev);
    }
  }
  initframe = frame - 1;

/*
  for(quitframe = nframe - 1; energy[quitframe] < elimit; quitframe--);

  printf("%d\t%d\n", framestep * initframe, framestep * quitframe);
*/

  printf("%d\n", framestep * initframe);

  return(0);
}
