/******************************************************************************
*                                                                             *
*       Copyright (C) 1993,1994 Tony Robinson				      *
*                                                                             *
*       See the file SLICENSE for conditions on distribution and usage	      *
*                                                                             *
******************************************************************************/

/*
  This is Sgetargs().

  The aim of this function is to provide hassle free command line
  parsing for most user applications.

  If you need tradition, use the standard getarg();
  If you need bells and whistles, use the GNU getargs();
  If you need to use non-standard syntax, write it yourself.

  This parser accepts the following syntax in argv, argc:

    argv[0] is assumed to contain the program name.
    The remainer of the line contains options then files.
    Options can either be single word boolean flags or two word
      (name, value) pairs.
    Files are either valid pathnames or '-', in which case a default
      is used.

  See Sutil/argtest.c for an example of use.

*/

# include <sys/types.h>
# include <stdio.h>
# include <stdlib.h>
# include <stdarg.h>
# include "Slib.h"

/* a linked list for the help flags */
typedef struct helpl_struct {
  char* flg;	/* a help flag */
  struct helpl_struct* nxt;
} helpl_type;

/* a linked list for the command line options */
typedef struct optl_struct {
  char* flg;	/* the optional command line argument */
  char* fmt;	/* the printf format of the value of the option */
  char  typ;    /* the scanf format of the value of the option */
  void* val;	/* a pointer to the storage location of the default value */
  char* msg;	/* a message explaining the usage of this option */
  struct optl_struct* nxt;
} optl_type;

/* a linked list for the command line files */
typedef struct filel_struct {
  char*  flg;	/* the generic file type */
  char*  mod;	/* the mode in which the file is to be opened */
  FILE** fil;	/* the pointer to the default file and returned file */
  char** nam;	/* the pointer to the name of the returned file */
  char*  msg;	/* a message explaining the usage of this file */
  struct filel_struct* nxt;
} filel_type;

/* checks a flag with a list of help flags and returns the match */
int checkHelp(char* flag, helpl_type *helpl) {
  int match = 0;
  for(match = 0; helpl != NULL && !match; helpl = helpl->nxt)
    match = (strcmp(flag, helpl->flg) == 0);
  return(match);
}

/* print the data at *val with type typ according to format fmt */
void fprintfptr(FILE* stream, char* fmt, char typ, void* val) {
  if(typ == 'c')
    fprintf(stream, fmt, *((char*) val));
  else if(typ == 's')
    if(*(char**) val != NULL)
      fprintf(stream, fmt, *((char**) val));
    else
      fprintf(stream, "(null)");
  else if(typ == 'd')
    fprintf(stream, fmt, *((int*) val));
  else if(typ == 'f')
    fprintf(stream, fmt, *((float*) val));
  else {
    fprintf(stderr, "\nSgetargs: format not recognised: %s\n", fmt);
    fprintf(stderr, "Sgetargs: please write me!\n");
    exit(1);
  }
}

/* scan a string into *val with type typ according to format fmt */
void sscanfptr(char* string, char* fmt, char typ, void* val) {
  if(typ == 'c')
    sscanf(string, "%c", ((char*) val));
  else if(typ == 's')
    *(char**) val = string;
  else if(typ == 'd')
    sscanf(string, "%d", ((int*) val));
  else if(typ == 'f')
    sscanf(string, "%f", ((float*) val));
  else {
    fprintf(stderr, "\nSgetargs: format not recognised: %s\n", fmt);
    fprintf(stderr, "Sgetargs: please write me!\n");
    exit(1);
  }
}

/* parse the command line argument in argv */
int Sgetargs(int argc, char** argv, ...) {
  va_list ap;
  char *flg;
  int nhelp, nopt, nfile, j, argfound = 1;
  helpl_type *helpl = NULL, **phelpl, *helpli;
  optl_type  *optl  = NULL, **poptl,  *optli;
  filel_type *filel = NULL, **pfilel, *fileli;

  va_start(ap, argv);

  /* read the list of help flags into a linked list */
  phelpl = &helpl;
  for(nhelp = 0; (flg = va_arg(ap, char*)) != NULL; nhelp++) {
    helpli = *phelpl = malloc(sizeof(helpl_type));
    helpli->flg = flg;
    helpli->nxt = NULL;
    phelpl = &(helpli->nxt);
  }

  /* read the options into a linked list */
  poptl = &optl;
  for(nopt = 0; (flg = va_arg(ap, char*)) != NULL; nopt++) {
    optli = *poptl = malloc(sizeof(optl_type));
    optli->flg = flg;
    optli->fmt = va_arg(ap, char*);

    /* crudely parse the printf format for a simple type */
    /* this must exist in the last two characters, and be l? or ? */
    if(optli->fmt != NULL) {
      optli->typ = *(optli->fmt + strlen(optli->fmt) - 1);
    }

    optli->val = va_arg(ap, void*);

    /* if a flag, then set the default to 0 */
    if(optli->fmt == NULL)
      *(int*) optli->val = 0;

    optli->msg = va_arg(ap, char*);
    optli->nxt = NULL;

    poptl = &(optli->nxt);
  }

  pfilel = &filel;
  for(nfile = 0; (flg = va_arg(ap, char*)) != NULL; nfile++) {
    fileli = *pfilel = malloc(sizeof(filel_type));
    fileli->flg = flg;
    fileli->mod = va_arg(ap, char*);
    fileli->fil = va_arg(ap, FILE**);
    fileli->nam = va_arg(ap, char**);
    fileli->msg = va_arg(ap, char*);
    fileli->nxt = NULL;

    pfilel = &(fileli->nxt);
  }
  va_end(ap);

  /* check for a help flag and print out the options and defaults if found */
  for(j = 1; j < argc; j++) {
    if(checkHelp(argv[j], helpl)) {
      int nprint;

      printf("%s: [options]", argv[0]);
      for(fileli = filel; fileli != NULL; fileli = fileli->nxt) {
	if(*(fileli->flg) == NULL)
	  printf(" %s", fileli->flg);
	else
	  printf(" [%s]", fileli->flg);
      }
      printf("\n");

      printf("    ");
      nprint = 0;
      for(helpli = helpl; helpli != NULL; helpli = helpli->nxt) {
	if(helpli->nxt != NULL)
	  printf("%s, ", helpli->flg);
	else
	  printf("or %s", helpli->flg);
      }
      printf("\tthis message\n");

      for(optli = optl; optli != NULL; optli = optli->nxt) {
	printf("    %-11s ", optli->flg);
	if(optli->fmt != NULL)
      	  fprintfptr(stdout, optli->fmt, optli->typ, optli->val);
        printf("\t%s\n", optli->msg);
      }

      for(fileli = filel; fileli != NULL; fileli = fileli->nxt) {
	if(*(fileli->flg) == NULL)
	  printf("    %s\t\t%s\n", fileli->flg, fileli->msg);
	else
	  printf("    [%s]\t\t%s\n", fileli->flg, fileli->msg);
      }
    exit(0);
    }
  }

  /* check for a match */
  for(j = 1; j < argc && argfound; j++) {
    argfound = 0;

    for(optli = optl; optli != NULL; optli = optli->nxt) {
      if(strcmp(argv[j], optli->flg) == 0) {
        argfound = 1;
        if(optli->fmt == NULL) /* found a flag, so set it to true */
          *(int*) optli->val = 1;
        else { /* scan in the next argument */
          j++;
	  if(j == argc) {
	    fprintf(stderr, "Sgetargs: truncated arg list\n");
	    fprintf(stderr, "%s: use %s for more information\n", argv[0],
		    helpl->flg);
	    exit(-1);
	  }
	  sscanfptr(argv[j], optli->fmt, optli->typ, optli->val);
        }
      }
    }
  }
  j--;

  if(j + nfile != argc) {
    fprintf(stderr, "%s: command line syntax error at: %s\n", argv[0],argv[j]);
    fprintf(stderr, "%s: use \"%s\" for more information\n", argv[0], helpl->flg);
    exit(-1);
  }

  /* parse the remaining arguments in the command line */
  for(fileli = filel; fileli != NULL; fileli = fileli->nxt, j++) {

    /* test to see if the defaults should be used */
    if(strcmp(argv[j], "-") == 0) {
      /* test to see if a default stream has been specified */
      if(fileli->fil == NULL) {
	fprintf(stderr, "%s: no default file specified\n", argv[0]);
        fprintf(stderr, "%s: use \"%s\" for more information\n", argv[0],
		helpl->flg);
        exit(-1);
      }
    }
    else {
      *(fileli->fil) = fopen(argv[j], fileli->mod);
      if(*(fileli->fil) == NULL) {
	fprintf(stderr, "%s: Sgetargs: fopen(\"%s\", \"%s\"): ", argv[0],
		argv[j], fileli->mod);
	perror(NULL);
	exit(-1);
      }
    }
    /* fill in the name of the file if required */
    if(fileli->nam != NULL)
      *(fileli->nam) = argv[j];
  }

  return(nopt);
}
