/******************************************************************************
*                                                                             *
*       Copyright (C) 1993 Tony Robinson				      *
*                                                                             *
*       See the file SLICENSE for conditions on distribution and usage	      *
*                                                                             *
******************************************************************************/

# include <math.h>
# include <stdio.h>
# include <sys/types.h>
# include "Slib.h"

/* find the first nacf autocorrelation values of the nwav samples in wav */

void Swav2acf(float *wav, int nwav, float *acf, int nacf) {
  int i, j;

  for(i = 0; i < nacf; i++) {
    double sum = 0.0;
    for(j = 0; j < nwav - i; j++)
      sum += wav[j] * wav[j + i];
    acf[i] = sum;
  }
}

/* find the lpc coefficents and reflection coefficients from the
   autocorrelation function */

float Sacf2lpc(float *acf, int nacf, float *lpc, float *ref) {
  static float *tmp;
  static int ntmp = 0;
  double e, ci;
  int i, j;

  if(ntmp != nacf) {
    if(ntmp != 0) Sfree(tmp);
    tmp = Svector(nacf, sizeof(*tmp));
    ntmp = nacf;
  }
  
  /* find lpc coefficients */
  e = acf[0];
  lpc[0] = 1.0;
  for(i = 1; i <= nacf; i++) {
    ci = 0.0;
    for(j = 1; j < i; j++) ci += lpc[j] * acf[i-j];
    ref[i] = ci = (acf[i] - ci) / e;
    lpc[i] = ci;
    for(j = 1; j < i; j++) tmp[j] = lpc[j] - ci * lpc[i-j];
    for(j = 1; j < i; j++) lpc[j] = tmp[j];
    e = (1 - ci * ci) * e;
  }
  return(e);
}

/* convert lpc coefficents to cepstral coefficients */

void Slpc2cep(float *lpc, int nlpc, float *cep, int ncep) {
  int n;
  
  for(n = 1; n <= nlpc && n <= ncep; n++) {
    float sum = 0.0;
    int   k;

    for(k = 1; k < n; k++) sum += k * cep[k] * lpc[n - k];
    cep[n] = lpc[n] + sum / n;
  }


  /* be wary of these interpolated values */
  for(n = nlpc + 1; n <= ncep; n++) {
    float sum = 0.0;
    int   k;

    for(k = n - (nlpc - 1); k < n; k++) sum += k * cep[k] * lpc[n - k];
    cep[n] = sum / n;
  }
}

void Swav2cep(float *wav, int nwav, float *cep, int ncep) {
  static int nint = 0;
  static float *acf = NULL, *lpc = NULL, *ref = NULL;

  if(nint != ncep) {
    if(acf != NULL) Sfree(acf);
    if(lpc != NULL) Sfree(lpc);
    if(ref != NULL) Sfree(ref);
    acf = Svector(ncep, sizeof(*acf));
    lpc = Svector(ncep, sizeof(*lpc));
    ref = Svector(ncep, sizeof(*ref));
    nint = ncep;
  }
  
  Swav2acf(wav, nwav, acf, ncep);
  Sacf2lpc(acf, ncep - 1, lpc, ref);
  cep[0] = log(acf[0]);
  Slpc2cep(lpc, ncep - 1, cep, ncep - 1);
}

void Sref2area(float *ref, float *area, int order) {
  int   i;

  for(i = 1; i <= order; i++)
    area[i] = (1.0 - ref[i]) / (1.0 + ref[i]);
}

#define ALMOST1 0.99999
void Sref2logarea(float *ref, float *logarea, int order) {
  int   i;

  for(i = 1; i <= order; i++) {
    if(ref[i] > ALMOST1) logarea[i] = log((1.0 - ALMOST1) / (1.0 + ALMOST1));
    else if(ref[i] < -ALMOST1) logarea[i] = log((1.0 + ALMOST1)/(1.0-ALMOST1));
    else logarea[i] = log((1.0 - ref[i]) / (1.0 + ref[i]));
  }
}

void Sref2truearea(float *ref, float *area, int order) {
  int   i;

  area[1] = (1.0 - ref[1]) / (1.0 + ref[1]);
  for(i = 2; i <= order; i++) 
    area[i] = area[i - 1] * (1.0 - ref[i]) / (1.0 + ref[i]);
}

#ifdef CLEAN_UP_THIS_CODE

int check_qlpc(qlpc, nqlpc) float *qlpc; int nlpc; {
  int j, k, stable = 1;
  float tmpref[MAX_LPC_ORDER];

  /* check for stabilty */
  for(j = 1; j <= nqlpc; j++)
    tmpref[j] = qlpc[j] / (float) (1 << LPCQUANT);
      
  for(k = nqlpc; k > 0 && stable; k--) {
    for(j = k - 1; j > 0; j--)
      tmpref[j] = (tmpref[j] + tmpref[k] * tmpref[k - j]) /
	(1.0 - tmpref[k] * tmpref[k]);
    if(fabs(tmpref[k]) >= 1.0) stable = 0;
  }	
  return(stable);
}

/******************************************************************************
The following is taken from:

"Low Bit rate Quantisation of the Smoothed Group Delay Spectrum for Speech
Recognition", Harald SINGER, Taizo UMEZAKI and Fumitada ITAKURA, ICASSP-90,
volume 2, page 761.
******************************************************************************/

void lpc2sgds(float *lpc, int nlpc, float *sgds, int nsgds, float minf, float maxf, float samplef) {
  if(nsgds > 0) {
    static int    nceps = 0;
    static float *pceps = NULL;
    float  minchan = Hz2Nonlinear(minf), maxchan = Hz2Nonlinear(maxf);
    float  chanstep = (maxchan - minchan) / (float) nsgds;
    int    i, k;

    if(nceps != nlpc) {
      if(nceps != 0) Sfree((char*) pceps);
      pceps = (float*) Smalloc(nlpc * sizeof(*pceps));
      nceps = nlpc;
    }

    /*  Check this line before using it! */
    lpc2ceps(lpc, nlpc, pceps - 1, nceps);

    for(i = 0; i < nsgds; i++) {
      float cfreq = 2.0 * M_PI * Nonlinear2Hz((i + 0.5) * chanstep + minchan) /
                 samplef;
      float hband = M_PI * (Nonlinear2Hz(i * chanstep + minchan) - 
		 Nonlinear2Hz((i + 1) * chanstep + minchan))/samplef;
      float sum = 0.0;

      for(k = 0; k < nceps; k++) 
        sum += sin(k * hband) * pceps[k] * cos(k * cfreq);
      sgds[i] = sum / hband;
    }
  }
}


void lpc2chan(float *lpc, int nlpc, float *chan, int nchan, float minf, float maxf, float samplef) {
  float  scale, total = 1.0e-36;
  float  minchan = Hz2Nonlinear(minf), maxchan = Hz2Nonlinear(maxf);
  float  chanstep = (maxchan - minchan) / (float) nchan;
  int    i, j;

  for(i = 0; i < nchan; i++) {
    float sums = 0.0, sumc = 1.0;
    float freq = 2.0 * M_PI * Nonlinear2Hz((i + 0.5)*chanstep+minchan)/samplef;

    for(j = 1; j <= nlpc; j++) {
      sums -= sin(freq * j) * lpc[j];
      sumc -= cos(freq * j) * lpc[j];
    }
    total += (chan[i] = sums * sums + sumc * sumc);
  }

  scale = nchan / total;
  for(i = 0; i < nchan; i++) chan[i] *= scale;
}

#endif
