/******************************************************************************
*                                                                             *
*       Copyright (C) 1993,1994 Tony Robinson				      *
*                                                                             *
*       See the file SLICENSE for conditions on distribution and usage	      *
*                                                                             *
******************************************************************************/

# include <sys/types.h>
# include <stdio.h>
# include <math.h>
# include <stdlib.h>
# include "Slib.h"

/* compilation options */
#define CEPSTRA
#define NODELTAS

# define DEFAULT_HEADERSIZE	1024
# define DEFAULT_FRAMESTEP	256
# define DEFAULT_WINDOWSIZE	(2 * DEFAULT_FRAMESTEP)
# define DEFAULT_NCEPSTRA	12
# define DEFAULT_MINSILENCE	4
# define DEFAULT_VQFF		1.0


int main(argc, argv) int argc; char **argv; {
  int swapbytes;
  int headersize   = DEFAULT_HEADERSIZE;
  int framestep    = DEFAULT_FRAMESTEP;
  int windowsize   = DEFAULT_WINDOWSIZE;
  int ncepstra     = DEFAULT_NCEPSTRA;
  int minsilence   = DEFAULT_MINSILENCE;
  float vqff	   = DEFAULT_VQFF;
  FILE *filei = NULL;
  short *wave;
  float *window;
  float **cepCache, *silMean, *sigMean, *silVar, *sigVar;
  double *silSum, *sigSum, *silSumSq, *sigSumSq;
  int   nsample, frame, nframe;
  int	i, ninit, nquit, nacoustic, lastninit, lastnquit;

  Sgetargs(argc, argv, "-h", "-help", NULL,
	"-swapbytes",	NULL, &swapbytes, "swap the byte order of the signals",
	"-headersize",	"%d", &headersize, "size of header in bytes",
	"-framestep",	"%d", &framestep, "analysis frame step",
	"-windowsize",	"%d", &windowsize, "analysis window size",
	"-ncepstra",  "%d",   &ncepstra, "number of cepstral coefficients",
	"-minsilence",  "%d", &minsilence, "minimum number of silence frames as each end",
	"-vqff",  "%4.2f",    &vqff, "Vector Quantisation Fudge Factor",
	NULL,
	"speech file", "r", &filei, NULL, "16 bit speech waveform file",
	NULL);

  /* read in the waveform */
  wave = Sfreadwave(filei, NULL, headersize, windowsize - framestep, &nsample);
  if(swapbytes)
    Sswab(wave, wave, 2 * (nsample + windowsize - framestep));

#ifdef DELTAS
  nacoustic = 2 * ncepstra;
#else
  nacoustic = ncepstra;
#endif

  nframe = nsample / framestep;
  window = Svector(windowsize, sizeof(*window));
  cepCache = (float**) Smatrix(nframe, nacoustic, sizeof(**cepCache));

  /* compute all the cepstra */
  for(frame = 0; frame < nframe; frame++) {
    SzeroMeanHammWindow(wave + frame * framestep, windowsize, window);
#ifdef CEPSTRA
    Swav2cep(window, windowsize, cepCache[frame], ncepstra);
#else
    Swav2acf(window, windowsize, cepCache[frame], ncepstra);
    for(i = 1; i < ncepstra; i++)
      cepCache[frame][i] /= cepCache[frame][0];
    cepCache[frame][0] = pow(cepCache[frame][0], 0.333333);
#endif
  }


#ifdef DELTAS
  /* add the deltas */
  for(i = 0; i < ncepstra; i++)
    cepCache[0][i + ncepstra] = 2 * (cepCache[1][i] - cepCache[0][i]);
  for(frame = 1; frame < nframe - 1; frame++)
    for(i = 0; i < ncepstra; i++)
      cepCache[frame][i + ncepstra] = 
	cepCache[frame + 1][i] - cepCache[frame - 1][i];
  for(i = 0; i < ncepstra; i++)
    cepCache[nframe - 1][i + ncepstra] =
      2 * (cepCache[nframe - 1][i] - cepCache[nframe - 2][i]);
#endif

  /* scale every component to have unity variance */
  for(i = 0; i < nacoustic; i++) {
    float sumx = 0.0;
    float sumxx = 0.0;
    float mean, var;

    for(frame = 0; frame < nframe; frame++) {
      float tmp = cepCache[frame][i];
      sumx  += tmp;
      sumxx += tmp * tmp;
    }
    mean = sumx / nframe;
    var  = sumxx / nframe - mean * mean;

    for(frame = 0; frame < nframe; frame++)
      cepCache[frame][i] /= var;
  }

  /* grab space for silence and signal sums and means */
  silSum   = Svector(nacoustic, sizeof(*silSum));
  sigSum   = Svector(nacoustic, sizeof(*sigSum));
  silMean  = Svector(nacoustic, sizeof(*silMean));
  sigMean  = Svector(nacoustic, sizeof(*sigMean));
  silSumSq = Svector(nacoustic, sizeof(*silSumSq));
  sigSumSq = Svector(nacoustic, sizeof(*sigSumSq));
  silVar   = Svector(nacoustic, sizeof(*silVar));
  sigVar   = Svector(nacoustic, sizeof(*sigVar));

  /* assume that the first and the last frames are silence */
  ninit = minsilence;
  nquit = minsilence;

  /* compute the sum of the silence vectors */
  for(i = 0; i < nacoustic; i++) {
    silSum[i]   = 0.0;
    silSumSq[i] = 0.0;

    for(frame = 0; frame < ninit; frame++) {
      silSum[i]   += cepCache[frame][i];
      silSumSq[i] += cepCache[frame][i] * cepCache[frame][i];
    }
    
    for(frame = nframe - nquit; frame < nframe; frame++) {
      silSum[i]   += cepCache[frame][i];
      silSumSq[i] += cepCache[frame][i] * cepCache[frame][i];
    }
  }
      
  /* compute the sum of the signal vectors */
  for(i = 0; i < nacoustic; i++) {
    sigSum[i]   = 0.0;
    sigSumSq[i] = 0.0;

    for(frame = ninit; frame < nframe - nquit; frame++) {
      sigSum[i]   += cepCache[frame][i];
      sigSumSq[i] += cepCache[frame][i] * cepCache[frame][i];
    }
  }
  
  lastninit = 0;
  lastnquit = 0;

  /* iterate to convergence */
  while(ninit != lastninit || nquit != lastnquit) {
    int maxsig;

    /* store the last boundaries so we can see if they change */
    lastninit = ninit;
    lastnquit = nquit;

    /* compute means and variances of silence and signal */
    for(i = 0; i < nacoustic; i++) {
      int nsil = ninit + nquit;
      int nsig = nframe - nsil;

      silMean[i] = silSum[i] / nsil;
      silVar[i]  = silSumSq[i] / nsil - silMean[i] * silMean[i];

      sigMean[i] = sigSum[i] / nsig;
      sigVar[i]  = sigSumSq[i] / nsig - sigMean[i] * sigMean[i];

      /* CHEAT! */
      silVar[i] = (1.0 - vqff) * silVar[i] + vqff * sigVar[i];
    }

    /* see if the start point can be increased */
    if(Sllikediag(cepCache[ninit], silMean, silVar, nacoustic) >
       Sllikediag(cepCache[ninit], sigMean, sigVar, nacoustic)) {
      for(i = 0; i < nacoustic; i++) {
	silSum[i]   += cepCache[ninit][i];
	sigSum[i]   -= cepCache[ninit][i];
	silSumSq[i] += cepCache[ninit][i] * cepCache[ninit][i];
	sigSumSq[i] -= cepCache[ninit][i] * cepCache[ninit][i];
      }
      ninit++;
    }

    /* see if the end point can be decreased */
    maxsig = nframe - nquit - 1;
    if(Sllikediag(cepCache[maxsig], silMean, silVar, nacoustic) >
       Sllikediag(cepCache[maxsig], sigMean, sigVar, nacoustic)) {
      for(i = 0; i < nacoustic; i++) {
	silSum[i]   += cepCache[maxsig][i];
	sigSum[i]   -= cepCache[maxsig][i];
	silSumSq[i] += cepCache[maxsig][i] * cepCache[maxsig][i];
	sigSumSq[i] -= cepCache[maxsig][i] * cepCache[maxsig][i];
      }
      nquit++;
    }
  }

  printf("%d\t%d\n", ninit * framestep, (nframe - nquit) * framestep);

  return(0);
}
