/* This program is an algorithm implementation of
   "Suppression of Acoustic Noise in Speech Using Spectral Subtraction"
   by Steven F. Boll.  For reference, please read
   "IEEE Transactions on Acoustics, Speech, and Signal Processing,
   Vol. ASSP-27, No.2, April 1979".

   The main steps in the noise suppression procedure are bias removal,
   half-wave rectification, residual noise reduction & nonspeech attenuation.

   Written by Richard Lee

   Modified by Tony Robinson

   Work is needed to bring it in line with the current CookBook tools.
   In particular, the call to realft should be replaced by a freely
   redistributable FFT program.
*/

# include <sys/types.h>
# include <stdio.h>
# include <math.h>
# include <stdlib.h>
# include <string.h>
# include "Slib.h"

/* abort if any math errors are found */
int matherr(x)
register struct exception *x;
{
  abort( );
}

# define DEFAULT_HEADERSIZE	94	/* headersize of aiff files */
# define DEFAULT_FRAMESTEP	128	/* integer power of 2 for FFT */
# define DEFAULT_WINDOWSIZE	(2 * DEFAULT_FRAMESTEP)
# define DEFAULT_NOSPEECH	12.0	/* average power ratio (in dB) (signal
				after residual noise reduction vs. mean noise
				noise magnitude) classified as no speech */
# define DEFAULT_ATTENUATION	30.0	/* output signal attenuation during
					   non-speech activity (in dB) */
# define ADDSN	2.0	/* At present the function Sarggetlf does not work
			   properly, & noise is always added to the signal in
			   the multiple ADDSN */

#ifndef M_PI
#define M_PI		3.14159265358979323846
#endif

void  realft(float*, int, int);
float mod_cmx(float*);
int   framestep  = DEFAULT_FRAMESTEP;
int   windowsize = DEFAULT_WINDOWSIZE;
float *hammingwindow;

int main(int argc, char **argv) {
  void  proc_noise(short*, float*, float*, int);
  float min_absreal(float**, int);
  float min_modcmx(float**, int);
  int   headersize = DEFAULT_HEADERSIZE;
  float nospeech   = DEFAULT_NOSPEECH;
  float atten      = DEFAULT_ATTENUATION;
  char  *file_signal, *file_noise;
  FILE  *fileo;
  short *signal, *noise, *sh_proc;
  float *processed;
  float *sn, *sp, *hwaves, **hwaveori, **hwavernr;
  float addsn =  ADDSN;
  float sum, mean, *meanmodn, *maxmodn, filter, mod, mods_factor;
  int   swapbytes, nwindows, nwindown, nsamples, nsamplen;
  int   i, window, ori_c, ori_p, rnr_c, rnr_p, hsample;

  /* this handles arguments in the Unix command line */
  Sgetargs(argc, argv, "-?", "-help", NULL,
	   "-a", "%6.4f", &addsn, "noise_multiple\tAdd background noise digitally to\n\t\t\tspeech before any processing (useful in debugging\n\t\t\tif the original speech is not degraded by noise)\n\t\t\tNoise amplitudes are multiplied by the (non-zero)\n\t\t\tnoise_multiple given after \"-a\" before addition",
	   "-h", "%d", &headersize, "headersize",
	   "-s", NULL, &swapbytes, "Swap high & low bits  (for machines that store\n\t\t\thigh bits before low bits in a two byte integer)",
	   NULL,
	   "signal", "r", NULL, &file_signal, "Signal waveform file",
	   "noise",  "r", NULL, &file_noise,  "Noise waveform file",
	   "output", "w", &fileo, NULL,  "Spectrally subtracted waveform file",
	   NULL);

  if(swapbytes)
    Spanic("%s: Sorry, option -swapbytes is not implemented!\n", *argv);

  /* set up the hamming window function*/
  hammingwindow = (float*) Svector(windowsize, sizeof(*hammingwindow));
  for (i = 0; i < windowsize; i++)
    hammingwindow[i] = 0.54 - 0.46 * cos(2.0 * M_PI * (i+1) / (windowsize-1));

  /* read noise_waveform_file */
  noise = (short*) Sreadshortfile(file_noise, headersize,
				  framestep, &nsamplen);

  /* ignore sample values which are not going to be windowed */
  nsamplen = nsamplen - nsamplen % framestep;
  nwindown = nsamplen / framestep - 1;

  /* allocate & initialise the space for storage of average and maximum noise
     magnitudes.  note that for odd i's except i=1, meanmodn[i] & maxmodn[i]
     are not used & are not initialised. */
  meanmodn = (float*) Svector(windowsize, sizeof(*meanmodn));
  meanmodn[1] = 0.0;
  for (i = 0; i < windowsize; i += 2)
    meanmodn[i] = 0.0;
  maxmodn = (float*) Svector(windowsize, sizeof(*maxmodn));
  maxmodn[1] = 0.0;
  for (i = 0; i < windowsize; i += 2)
    maxmodn[i] = 0.0;

  /* process noise_waveform_file */
  proc_noise(noise, meanmodn, maxmodn, nwindown);

  /* read signal waveform file */
  signal = (short*) Sreadshortfile(file_signal, headersize,
				   framestep, &nsamples);

  /* ignore signal sample values which are not going to be windowed */
  nsamples = nsamples - nsamples % framestep;

  /* add noise digitally to speech signal */
  sn = (float*) Svector(nsamples, sizeof(*sn));
  if (addsn != 0.0) {
    /* ignore sample values which have no noise values to add to */
    if (nsamples > nsamplen) {
      fprintf(stderr, "Not enough noise sample values\n");
      fprintf(stderr, "Signal sample values at the end discarded\n");
      nsamples = nsamplen;
    }
    for (i = 0; i < nsamples; i++)
      sn[i] = (signal[i] + addsn * noise[i])/(1.0 + addsn);
    addsn = addsn/(1.0 + addsn);	/* addsn is redefined */
    meanmodn[1] *= addsn;
    for (i = 0; i < windowsize; i += 2)
      meanmodn[i] *= addsn;
    maxmodn[1] *= addsn;
    for (i = 0; i < windowsize; i += 2)
      maxmodn[i] *= addsn;
  }
  else
    for (i = 0; i < nsamples; i++)
      sn[i] = signal[i];

  /* allocate space for hamming window */
  nwindows = nsamples / framestep - 1;
  hwaves = (float*) Svector(windowsize, sizeof(*hwaves));
  /* allocate space to store two adjacent hamming window frames for use in
     signal attenuation during non-speech activity (after residual noise
     reduction) */
  hwaveori = (float**) Smatrix(2, windowsize, sizeof(**hwaveori));
  /* allocate space to store three adjacent hamming window frames needed for
     residual noise reduction (or three frame averaging) */
  hwavernr = (float**) Smatrix(3, windowsize, sizeof(**hwavernr));
  /* allocate & initialise the space for the processed waveform */
  processed = (float*) Svector(nsamples, sizeof(*processed));
  for (i = 0; i < nsamples; i++)
    processed[i] = 0.0;

  /* hamming window & noise suppression */
  for (window = 0; window < nwindows; window++) {
    /* set pointer */
    sp = sn + window * framestep;

    /* window signal with mean set at zero */
    for (i = 0, sum = 0.0; i < windowsize; i++)
      sum += sp[i];
    mean = sum / windowsize;
    for (hsample = 0; hsample < windowsize; hsample++)
      hwaves[hsample] = (sp[hsample] - mean) * hammingwindow[hsample];

    /* FFT */
    realft(hwaves-1, windowsize, 1);

    /* save a (original) copy of hwaves just after FFT for use in signal
       attenuation during non-speech activity */
    ori_c = window % 2;	/* set current hamming window frame */
    ori_p = (window + 1) % 2;	/* set previous hamming window frame */
    for (hsample = 0; hsample < windowsize; hsample++)
      hwaveori[ori_c][hsample] = hwaves[hsample];

    /* subtract spectral bias with half-wave rectification */
    for (hsample = 0; hsample < 2; hsample++)
      if ((mod = (float) fabs(hwaves[hsample])) > meanmodn[hsample]) {
	filter = 1.0 - meanmodn[hsample] / mod;
	hwaves[hsample] *= filter;
      }
      else
	hwaves[hsample] = 0.0;
    for (hsample = 2; hsample < windowsize; hsample += 2)
      if ((mod = mod_cmx(hwaves + hsample)) > meanmodn[hsample]) {
	filter = 1.0 - meanmodn[hsample] / mod;
	hwaves[hsample] *= filter;
	hwaves[hsample+1] *= filter;
      }
      else
	hwaves[hsample] = hwaves[hsample+1] = 0.0;

    /* reduce noise residual */
    rnr_c = window % 3;	/* set current hamming window frame */
    rnr_p = (window + 2) % 3;	/* set previous hamming window frame */
    /* store data in hwavernr before data in hwaves get overwritten */
    for (hsample = 0; hsample < windowsize; hsample++)
      hwavernr[rnr_c][hsample] = hwaves[hsample];
    /* the main loop starts here: note that the previous window elements,
       rather than the current ones, are processed (specifically, hwaves will
       store data of the previous window) */
    if (window > 1) {
      for (hsample = 0; hsample < 2; hsample++)
	if ((mod = (float) fabs(hwavernr[rnr_p][hsample])) >= maxmodn[hsample])
	  hwaves[hsample] = hwavernr[rnr_p][hsample];
	else if (mod != 0.0) {
	  mods_factor = min_absreal(hwavernr, hsample) / mod;
	  hwaves[hsample] = mods_factor * hwavernr[rnr_p][hsample];
	}
	else
	  hwaves[hsample] = 0.0;
      for (hsample = 2; hsample < windowsize; hsample += 2)
	if ((mod = mod_cmx(hwavernr[rnr_p] + hsample)) >= maxmodn[hsample]) {
	  hwaves[hsample] = hwavernr[rnr_p][hsample];
	  hwaves[hsample+1] = hwavernr[rnr_p][hsample+1];
	}
	else if (mod != 0.0) {
	  mods_factor = min_modcmx(hwavernr, hsample) / mod;
	  hwaves[hsample] = mods_factor * hwavernr[rnr_p][hsample];
	  hwaves[hsample+1] = mods_factor * hwavernr[rnr_p][hsample+1];
	}
	else
	  hwaves[hsample] = hwaves[hsample+1] = 0.0;
    }
    else if (window == 1) {
      for (hsample = 0; hsample < 2; hsample++)
	if ((mod = (float) fabs(hwavernr[0][hsample])) >= maxmodn[hsample])
	  hwaves[hsample] = hwavernr[0][hsample];
	else if (mod != 0.0) {
	  mods_factor = (mod < fabs(hwavernr[1][hsample]))
	    ? mod : (float) fabs(hwavernr[1][hsample]);
	  mods_factor /= mod;
	  hwaves[hsample] = mods_factor * hwavernr[0][hsample];
	}
	else
	  hwaves[hsample] = 0.0;
      for (hsample = 2; hsample < windowsize; hsample += 2)
	if ((mod = mod_cmx(hwavernr[0] + hsample)) >= maxmodn[hsample]) {
	  hwaves[hsample] = hwavernr[0][hsample];
	  hwaves[hsample+1] = hwavernr[0][hsample+1];
	}
	else if (mod != 0.0) {
	  mods_factor = (mod < mod_cmx(hwavernr[1] + hsample))
	    ? mod : mod_cmx(hwavernr[1] + hsample);
	  mods_factor /= mod;
	  hwaves[hsample] = mods_factor * hwavernr[0][hsample];
	  hwaves[hsample+1] = mods_factor * hwavernr[0][hsample+1];
	}
	else
	  hwaves[hsample] = hwaves[hsample+1] = 0.0;
    }

    if (window > 0) {
      /* attenuate signal during non-speech activity */
      for (i = 2, sum = 0.0; i < windowsize; i += 2)
	sum += mod_cmx(hwaves + i) / meanmodn[i];
      sum = 2.0 * sum + (float) fabs(hwaves[0]/meanmodn[0])
		+ (float) fabs(hwaves[1]/meanmodn[1]);
      if (sum / windowsize <= pow(10.0, nospeech/-20.0))
	for (hsample = 0; hsample < windowsize; hsample++)
	  hwaves[hsample] = (float) (pow(10.0, (-1 * atten / 20.0)))
				* hwaveori[ori_p][hsample];

      /* inverse FFT (with every value multiplied by (windowsize/2)) */
      realft(hwaves-1, windowsize, -1);

      /* overlap & add results for the waveform */
      for (hsample = 0; hsample < windowsize; hsample++)
	processed[hsample + (window-1) * framestep] += hwaves[hsample];
    }
  }

  /* process the final window (this has not been processed since the noise
     residual reduction routine always processes the previous window) */
  /* reduce noise residual */
  for (hsample = 0; hsample < 2; hsample++)
    if ((mod = (float) fabs(hwavernr[rnr_c][hsample])) >= maxmodn[hsample])
      hwaves[hsample] = hwavernr[rnr_c][hsample];
    else if (mod != 0.0) {
      mods_factor = (fabs(hwavernr[rnr_p][hsample]) < mod)
	? (float) fabs(hwavernr[rnr_p][hsample]) : mod;
      mods_factor /= mod;
      hwaves[hsample] = mods_factor * hwavernr[rnr_c][hsample];
    }
    else
      hwaves[hsample] = 0.0;
  for (hsample = 2; hsample < windowsize; hsample += 2)
    if ((mod = mod_cmx(hwavernr[rnr_c] + hsample)) >= maxmodn[hsample]) {
      hwaves[hsample] = hwavernr[rnr_c][hsample];
      hwaves[hsample+1] = hwavernr[rnr_c][hsample];
    }
    else if (mod != 0.0) {
      mods_factor = (mod_cmx(hwavernr[rnr_p] + hsample) < mod)
	? mod_cmx(hwavernr[rnr_p] + hsample) : mod;
      mods_factor /= mod;
      hwaves[hsample] = mods_factor * hwavernr[rnr_c][hsample];
      hwaves[hsample+1] = mods_factor * hwavernr[rnr_c][hsample+1];
    }
    else
      hwaves[hsample] = hwaves[hsample+1] = 0.0;
  /* attenuate signal during non-speech activity */
  for (i = 2, sum = 0.0; i < windowsize; i += 2)
    sum += mod_cmx(hwaves + i) / meanmodn[i];
  sum = 2.0 * sum + (float) fabs(hwaves[0]/meanmodn[0])
		+ (float) fabs(hwaves[1]/meanmodn[1]);
  if (sum / windowsize <= pow(10.0, nospeech/-20.0))
    for (hsample = 0; hsample < windowsize; hsample++)
      hwaves[hsample] = (float) (pow(10.0, (-1 * atten / 20.0)))
				* hwaveori[ori_c][hsample];
  /* inverse FFT (with every value multiplied by (windowsize/2)) */
  realft(hwaves-1, windowsize, -1);
  /* overlap & add results for the waveform */
  for (hsample = 0; hsample < windowsize; hsample++)
    processed[hsample + (nwindows-1) * framestep] += hwaves[hsample];

  /* write out the new waveform */
  sh_proc = (short*) Svector((nsamples -= 2 * framestep), sizeof(*sh_proc));
  for (i = 0; i < nsamples; i++)
    sh_proc[i] = floor(processed[i + framestep]/(windowsize/2)/1.08 + 0.5);
  fwrite(sh_proc, sizeof(*sh_proc), nsamples, fileo);

  /* free the space for the waveform */
  Sfree(signal);
  Sfree(noise);
  
  return(0);
}



/* process the noise array.  it returns average & maximum noise magnitudes at
   different frequencies. */
void proc_noise(short *noise, float *meanmodn, float *maxmodn, int nwindown) {
  short *np;
  float *hwaven;
  float sum, mean, *modn;
  int   i, window, hsample;

  /* allocate space for hamming window */
  hwaven = (float*) Svector(windowsize, sizeof(*hwaven));
  /* allocate space for storage of noise magnitudes.  note that for odd i's
     except i=1, modn[i] are not used. */
  modn =(float*) Svector(windowsize, sizeof(*modn));

  /* hamming window & find average & maximum noise magnitudes */
  for (window = 0; window < nwindown; window++) {
    /* set pointer */
    np = noise + window * framestep;

    /* window noise with mean set at zero */
    for (i = 0, sum = 0.0; i < windowsize; i++)
      sum += np[i];
    mean = sum / windowsize;
    for (hsample = 0; hsample < windowsize; hsample++)
      hwaven[hsample] = (np[hsample] - mean) * hammingwindow[hsample];

    /* FFT */
    realft(hwaven-1, windowsize, 1);

    /* compute total noise magnitude at each frequency */
    meanmodn[0] += (modn[0] = (float) fabs(hwaven[0]));
    meanmodn[1] += (modn[1] = (float) fabs(hwaven[1]));
    for (i = 2; i < windowsize; i += 2)
      meanmodn[i] += (modn[i] = mod_cmx(hwaven + i));

    /* compute maximum noise magnitude at each frequency */
    if (maxmodn[1] < modn[1])
      maxmodn[1] = modn[1];
    for (i = 0; i < windowsize; i += 2)
      if (maxmodn[i] < modn[i])
	maxmodn[i] = modn[i];
  }

  /* compute average noise magnitude at each frequency */
  meanmodn[1] /= nwindown;
  for (i = 0; i < windowsize; i += 2)
    meanmodn[i] /= nwindown;
}



/* compute the modulus of complex number where the real & imaginary parts are
   the first two elements of the array cmx */
float mod_cmx(float *cmx) {
  return (float) sqrt(cmx[0] * cmx[0] + cmx[1] * cmx[1]);
}



/* find the minimum of the absolute values of (three) corresponding elements in
   three arrays, which are themselves pointers of data */
float min_absreal(float **data, int sample) {
  float a = (float) fabs(data[0][sample]);
  float b = (float) fabs(data[1][sample]);
  float c = (float) fabs(data[2][sample]);
  float z;

  z = (a < b) ? a : b;
  return (z < c) ? z : c;
}



/* find the minimum of the moduli of (three) corresponding complex numbers in
   three arrays, where the real & imaginary parts alternate & the arrays are
   themselves pointers of data */
float min_modcmx(float **data, int sample) {
  float a = mod_cmx(data[0] + sample);
  float b = mod_cmx(data[1] + sample);
  float c = mod_cmx(data[2] + sample);
  float z;

  z = (a < b) ? a : b;
  return (z < c) ? z : c;
}

#define SWAP(a,b) tempr=(a);(a)=(b);(b)=tempr

void four1(data,nn,isign)
float data[];
int nn,isign;
{
	int n,mmax,m,j,istep,i;
	double wtemp,wr,wpr,wpi,wi,theta;
	float tempr,tempi;

	n=nn << 1;
	j=1;
	for (i=1;i<n;i+=2) {
		if (j > i) {
			SWAP(data[j],data[i]);
			SWAP(data[j+1],data[i+1]);
		}
		m=n >> 1;
		while (m >= 2 && j > m) {
			j -= m;
			m >>= 1;
		}
		j += m;
	}
	mmax=2;
	while (n > mmax) {
		istep=2*mmax;
		theta=6.28318530717959/(isign*mmax);
		wtemp=sin(0.5*theta);
		wpr = -2.0*wtemp*wtemp;
		wpi=sin(theta);
		wr=1.0;
		wi=0.0;
		for (m=1;m<mmax;m+=2) {
			for (i=m;i<=n;i+=istep) {
				j=i+mmax;
				tempr=wr*data[j]-wi*data[j+1];
				tempi=wr*data[j+1]+wi*data[j];
				data[j]=data[i]-tempr;
				data[j+1]=data[i+1]-tempi;
				data[i] += tempr;
				data[i+1] += tempi;
			}
			wr=(wtemp=wr)*wpr-wi*wpi+wr;
			wi=wi*wpr+wtemp*wpi+wi;
		}
		mmax=istep;
	}
}

#undef SWAP

void realft(data,n,isign)
float data[];
int n,isign;
{
	int i,i1,i2,i3,i4,n2p3;
	float c1=0.5,c2,h1r,h1i,h2r,h2i;
	double wr,wi,wpr,wpi,wtemp,theta;
	void four1();

	theta=3.141592653589793/(double) n;
	if (isign == 1) {
		c2 = -0.5;
		four1(data,n,1);
	} else {
		c2=0.5;
		theta = -theta;
	}
	wtemp=sin(0.5*theta);
	wpr = -2.0*wtemp*wtemp;
	wpi=sin(theta);
	wr=1.0+wpr;
	wi=wpi;
	n2p3=2*n+3;
	for (i=2;i<=n/2;i++) {
		i4=1+(i3=n2p3-(i2=1+(i1=i+i-1)));
		h1r=c1*(data[i1]+data[i3]);
		h1i=c1*(data[i2]-data[i4]);
		h2r = -c2*(data[i2]+data[i4]);
		h2i=c2*(data[i1]-data[i3]);
		data[i1]=h1r+wr*h2r-wi*h2i;
		data[i2]=h1i+wr*h2i+wi*h2r;
		data[i3]=h1r-wr*h2r+wi*h2i;
		data[i4] = -h1i+wr*h2i+wi*h2r;
		wr=(wtemp=wr)*wpr-wi*wpi+wr;
		wi=wi*wpr+wtemp*wpi+wi;
	}
	if (isign == 1) {
		data[1] = (h1r=data[1])+data[2];
		data[2] = h1r-data[2];
	} else {
		data[1]=c1*((h1r=data[1])+data[2]);
		data[2]=c1*(h1r-data[2]);
		four1(data,n,-1);
	}
}
